@@ -98,6 +98,7 @@
 #include "nest/route.h"
 #include "nest/protocol.h"
 #include "nest/iface.h"
+#include "nest/mpls.h"
 #include "lib/resource.h"
 #include "lib/event.h"
 #include "lib/timer.h"
@@ -117,7 +118,7 @@
 pool *rt_table_pool;
 
 static slab *rte_slab;
-static linpool *rte_update_pool;
+linpool *rte_update_pool;
 
 list routing_tables;
 
@@ -144,66 +145,6 @@ net_init_with_trie(struct fib *f, void *N)
     trie_add_prefix(tab->trie_new, n->n.addr, n->n.addr->pxlen, n->n.addr->pxlen);
 }
 
-static inline net *
-net_route_ip4_trie(rtable *t, const net_addr_ip4 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP4(t->trie, n0, n)
-  {
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_vpn4_trie(rtable *t, const net_addr_vpn4 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP4(t->trie, (const net_addr_ip4 *) n0, px)
-  {
-    net_addr_vpn4 n = NET_ADDR_VPN4(px.prefix, px.pxlen, n0->rd);
-
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_ip6_trie(rtable *t, const net_addr_ip6 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP6(t->trie, n0, n)
-  {
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_vpn6_trie(rtable *t, const net_addr_vpn6 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP6(t->trie, (const net_addr_ip6 *) n0, px)
-  {
-    net_addr_vpn6 n = NET_ADDR_VPN6(px.prefix, px.pxlen, n0->rd);
-
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
 static inline void *
 net_route_ip6_sadr_trie(rtable *t, const net_addr_ip6_sadr *n0)
 {
@@ -237,69 +178,6 @@ net_route_ip6_sadr_trie(rtable *t, const net_addr_ip6_sadr *n0)
   return NULL;
 }
 
-static inline net *
-net_route_ip4_fib(rtable *t, const net_addr_ip4 *n0)
-{
-  net_addr_ip4 n;
-  net_copy_ip4(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_vpn4_fib(rtable *t, const net_addr_vpn4 *n0)
-{
-  net_addr_vpn4 n;
-  net_copy_vpn4(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_ip6_fib(rtable *t, const net_addr_ip6 *n0)
-{
-  net_addr_ip6 n;
-  net_copy_ip6(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_vpn6_fib(rtable *t, const net_addr_vpn6 *n0)
-{
-  net_addr_vpn6 n;
-  net_copy_vpn6(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
 
 static inline void *
 net_route_ip6_sadr_fib(rtable *t, const net_addr_ip6_sadr *n0)
@@ -345,164 +223,58 @@ net *
 net_route(rtable *tab, const net_addr *n)
 {
   ASSERT(tab->addr_type == n->type);
+  net_addr_union *nu = SKIP_BACK(net_addr_union, n, n);
 
-  switch (n->type)
-  {
-  case NET_IP4:
-    if (tab->trie)
-      return net_route_ip4_trie(tab, (net_addr_ip4 *) n);
-    else
-      return net_route_ip4_fib (tab, (net_addr_ip4 *) n);
+#define TW(ipv, what) \
+  TRIE_WALK_TO_ROOT_IP##ipv(tab->trie, &(nu->ip##ipv), var) \
+  { what(ipv, var); } \
+  TRIE_WALK_TO_ROOT_END; return NULL;
 
-  case NET_VPN4:
-    if (tab->trie)
-      return net_route_vpn4_trie(tab, (net_addr_vpn4 *) n);
-    else
-      return net_route_vpn4_fib (tab, (net_addr_vpn4 *) n);
+#define FW(ipv, what) do { \
+  net_addr_union nuc; net_copy(&nuc.n, n); \
+  while (1) { \
+    what(ipv, nuc.ip##ipv); if (!nuc.n.pxlen) return NULL; \
+    nuc.n.pxlen--; ip##ipv##_clrbit(&nuc.ip##ipv.prefix, nuc.ip##ipv.pxlen); \
+  } \
+} while(0); return NULL;
 
-  case NET_IP6:
-    if (tab->trie)
-      return net_route_ip6_trie(tab, (net_addr_ip6 *) n);
-    else
-      return net_route_ip6_fib (tab, (net_addr_ip6 *) n);
+#define FVR_IP(ipv, var) \
+    net *r; if (r = net_find_valid(tab, (net_addr *) &var)) return r;
 
-  case NET_VPN6:
-    if (tab->trie)
-      return net_route_vpn6_trie(tab, (net_addr_vpn6 *) n);
-    else
-      return net_route_vpn6_fib (tab, (net_addr_vpn6 *) n);
+#define FVR_VPN(ipv, var) \
+    net_addr_vpn##ipv _var0 = NET_ADDR_VPN##ipv(var.prefix, var.pxlen, nu->vpn##ipv.rd); FVR_IP(ipv, _var0);
 
-  case NET_IP6_SADR:
-    if (tab->trie)
-      return net_route_ip6_sadr_trie(tab, (net_addr_ip6_sadr *) n);
-    else
-      return net_route_ip6_sadr_fib (tab, (net_addr_ip6_sadr *) n);
+  if (tab->trie)
+    switch (n->type) {
+      case NET_IP4:   TW(4, FVR_IP);
+      case NET_VPN4:  TW(4, FVR_VPN);
+      case NET_IP6:   TW(6, FVR_IP);
+      case NET_VPN6:  TW(6, FVR_VPN);
 
-  default:
-    return NULL;
-  }
-}
-
-
-static int
-net_roa_check_ip4_trie(rtable *tab, const net_addr_ip4 *px, u32 asn)
-{
-  int anything = 0;
-
-  TRIE_WALK_TO_ROOT_IP4(tab->trie, px, px0)
-  {
-    net_addr_roa4 roa0 = NET_ADDR_ROA4(px0.prefix, px0.pxlen, 0, 0);
-
-    struct fib_node *fn;
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)
-    {
-      net_addr_roa4 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa4(roa, &roa0) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
+      case NET_IP6_SADR:
+	return net_route_ip6_sadr_trie(tab, (net_addr_ip6_sadr *) n);
+      default:
+	return NULL;
     }
-  }
-  TRIE_WALK_TO_ROOT_END;
+  else
+    switch (n->type) {
+      case NET_IP4:   FW(4, FVR_IP);
+      case NET_VPN4:  FW(4, FVR_VPN);
+      case NET_IP6:   FW(6, FVR_IP);
+      case NET_VPN6:  FW(6, FVR_VPN);
 
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
-
-static int
-net_roa_check_ip4_fib(rtable *tab, const net_addr_ip4 *px, u32 asn)
-{
-  struct net_addr_roa4 n = NET_ADDR_ROA4(px->prefix, px->pxlen, 0, 0);
-  struct fib_node *fn;
-  int anything = 0;
-
-  while (1)
-  {
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &n); fn; fn = fn->next)
-    {
-      net_addr_roa4 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa4(roa, &n) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
+      case NET_IP6_SADR:
+	return net_route_ip6_sadr_fib (tab, (net_addr_ip6_sadr *) n);
+      default:
+	return NULL;
     }
 
-    if (n.pxlen == 0)
-      break;
-
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
+#undef TW
+#undef FW
+#undef FVR_IP
+#undef FVR_VPN
 }
 
-static int
-net_roa_check_ip6_trie(rtable *tab, const net_addr_ip6 *px, u32 asn)
-{
-  int anything = 0;
-
-  TRIE_WALK_TO_ROOT_IP6(tab->trie, px, px0)
-  {
-    net_addr_roa6 roa0 = NET_ADDR_ROA6(px0.prefix, px0.pxlen, 0, 0);
-
-    struct fib_node *fn;
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)
-    {
-      net_addr_roa6 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa6(roa, &roa0) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
-    }
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
-
-static int
-net_roa_check_ip6_fib(rtable *tab, const net_addr_ip6 *px, u32 asn)
-{
-  struct net_addr_roa6 n = NET_ADDR_ROA6(px->prefix, px->pxlen, 0, 0);
-  struct fib_node *fn;
-  int anything = 0;
-
-  while (1)
-  {
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &n); fn; fn = fn->next)
-    {
-      net_addr_roa6 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa6(roa, &n) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
-    }
-
-    if (n.pxlen == 0)
-      break;
-
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
 
 /**
  * roa_check - check validity of route origination in a ROA table
@@ -522,22 +294,181 @@ net_roa_check_ip6_fib(rtable *tab, const net_addr_ip6 *px, u32 asn)
 int
 net_roa_check(rtable *tab, const net_addr *n, u32 asn)
 {
+  net_addr_union *nu = SKIP_BACK(net_addr_union, n, n);
+  int anything = 0;
+  struct fib_node *fn;
+
+#define TW(ipv) do {								\
+  TRIE_WALK_TO_ROOT_IP##ipv(tab->trie, &(nu->ip##ipv), var) {			\
+    net_addr_roa##ipv roa0 = NET_ADDR_ROA##ipv(var.prefix, var.pxlen, 0, 0);	\
+    ROA_PARTIAL_CHECK(ipv);							\
+  } TRIE_WALK_TO_ROOT_END;							\
+  return anything ? ROA_INVALID : ROA_UNKNOWN;					\
+} while (0)
+
+#define FW(ipv) do {								\
+  net_addr_roa##ipv roa0 = NET_ADDR_ROA##ipv(nu->ip##ipv.prefix, nu->ip##ipv.pxlen, 0, 0);\
+  while (1) {									\
+    ROA_PARTIAL_CHECK(ipv);							\
+    if (roa0.pxlen == 0) break;							\
+    roa0.pxlen--; ip##ipv##_clrbit(&roa0.prefix, roa0.pxlen);			\
+  }										\
+} while (0)
+
+#define ROA_PARTIAL_CHECK(ipv) do {						\
+  for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)	\
+  {										\
+    net_addr_roa##ipv *roa = (void *) fn->addr;				\
+    net *r = fib_node_to_user(&tab->fib, fn);					\
+    if (net_equal_prefix_roa##ipv(roa, &roa0) && rte_is_valid(r->routes))	\
+    {										\
+      anything = 1;								\
+      if (asn && (roa->asn == asn) && (roa->max_pxlen >= nu->ip##ipv.pxlen))	\
+      return ROA_VALID;							\
+    }										\
+  }										\
+} while (0)
+
   if ((tab->addr_type == NET_ROA4) && (n->type == NET_IP4))
   {
-    if (tab->trie)
-      return net_roa_check_ip4_trie(tab, (const net_addr_ip4 *) n, asn);
-    else
-      return net_roa_check_ip4_fib (tab, (const net_addr_ip4 *) n, asn);
+    if (tab->trie)  TW(4);
+    else	    FW(4);
   }
   else if ((tab->addr_type == NET_ROA6) && (n->type == NET_IP6))
   {
-    if (tab->trie)
-      return net_roa_check_ip6_trie(tab, (const net_addr_ip6 *) n, asn);
-    else
-      return net_roa_check_ip6_fib (tab, (const net_addr_ip6 *) n, asn);
+    if (tab->trie)  TW(6);
+    else	    FW(6);
   }
-  else
-    return ROA_UNKNOWN;	/* Should not happen */
+
+  return anything ? ROA_INVALID : ROA_UNKNOWN;
+#undef ROA_PARTIAL_CHECK
+#undef TW
+#undef FW
+}
+
+/**
+ * aspa_check - check validity of AS Path in an ASPA table
+ * @tab: ASPA table
+ * @path: AS Path to check
+ *
+ * Implements draft-ietf-sidrops-aspa-verification-16.
+ */
+enum aspa_result aspa_check(rtable *tab, const adata *path, bool force_upstream)
+{
+  struct lp_state lps;
+  lp_save(tmp_linpool, &lps);
+
+  /* No support for confed paths */
+  if (as_path_contains_confed(path))
+    return ASPA_INVALID;
+
+  /* Check path length */
+  uint len = as_path_getlen(path);
+  if (len == 0)
+    return ASPA_INVALID;
+
+  /* Normalize the AS Path: drop stuffings */
+  u32 *asns = alloca(sizeof(u32) * len);
+  uint ppos = 0;
+  uint nsz = 0;
+  while (as_path_walk(path, &ppos, &asns[nsz]))
+    if ((nsz == 0) || (asns[nsz] != asns[nsz-1]))
+      nsz++;
+
+  /* Find the provider blocks for every AS on the path
+   * and check allowed directions */
+  uint max_up = 0, min_up = 0, max_down = 0, min_down = 0;
+
+  for (uint ap=0; ap<nsz; ap++)
+  {
+    net_addr_union nau = { .aspa = NET_ADDR_ASPA(asns[ap]), };
+    net *n = net_find(tab, &nau.n);
+
+    bool found = false, down = false, up = false;
+
+    for (rte *e = (n ? n->routes: NULL); e; e = e->next)
+    {
+      if (!rte_is_valid(e))
+	continue;
+
+      eattr *ea = ea_find(e->attrs->eattrs, EA_ASPA_PROVIDERS);
+      if (!ea)
+	continue;
+
+      /* Actually found some ASPA */
+      found = true;
+
+      for (uint i=0; i * sizeof(u32) < ea->u.ptr->length; i++)
+      {
+	if ((ap > 0) && ((u32 *) ea->u.ptr->data)[i] == asns[ap-1])
+	  up = true;
+	if ((ap + 1 < nsz) && ((u32 *) ea->u.ptr->data)[i] == asns[ap+1])
+	  down = true;
+
+	if (down && up)
+	  /* Both peers found */
+	  goto end_of_aspa;
+      }
+    }
+end_of_aspa:;
+
+    /* Fast path for the upstream check */
+    if (force_upstream)
+    {
+      if (!found)
+	/* Move min-upstream */
+	min_up = ap;
+      else if (ap && !up)
+	/* Exists but doesn't allow this upstream */
+	return ASPA_INVALID;
+    }
+
+    /* Fast path for no ASPA here */
+    else if (!found)
+    {
+      /* Extend max-downstream (min-downstream is stopped by unknown) */
+      max_down = ap+1;
+
+      /* Move min-upstream (can't include unknown) */
+      min_up = ap;
+    }
+
+    /* ASPA exists and downstream may be extended */
+    else if (down)
+    {
+      /* Extending max-downstream always */
+      max_down = ap+1;
+
+      /* Extending min-downstream unless unknown seen */
+      if (min_down == ap)
+	min_down = ap+1;
+
+      /* Downstream only */
+      if (!up)
+	min_up = max_up = ap;
+    }
+
+    /* No extension for downstream, force upstream only from now */
+    else
+    {
+      force_upstream = 1;
+
+      /* Not even upstream, move the ending here */
+      if (!up)
+	min_up = max_up = ap;
+    }
+  }
+
+  /* Is the path surely valid? */
+  if (min_up <= min_down)
+    return ASPA_VALID;
+
+  /* Is the path maybe valid? */
+  if (max_up <= max_down)
+    return ASPA_UNKNOWN;
+
+  /* Now there is surely a valley there. */
+  return ASPA_INVALID;
 }
 
 /**
@@ -675,7 +606,7 @@ rte_mergable(rte *pri, rte *sec)
 static void
 rte_trace(struct channel *c, rte *e, int dir, char *msg)
 {
-  log(L_TRACE "%s.%s %c %s %N %uL %uG %s",
+  log(L_TRACE "%s.%s %c %s %N %luL %uG %s",
       c->proto->name, c->name ?: "?", dir, msg, e->net->n.addr, e->src->private_id, e->src->global_id,
       rta_dest_name(e->attrs->dest));
 }
@@ -975,7 +906,6 @@ rt_export_merged(struct channel *c, net *net, rte **rt_free, linpool *pool, int
   return best;
 }
 
-
 static void
 rt_notify_merged(struct channel *c, net *net, rte *new_changed, rte *old_changed,
 		 rte *new_best, rte *old_best, int refeed)
@@ -1206,7 +1136,7 @@ rte_free_quick(rte *e)
   sl_free(e);
 }
 
-static int
+int
 rte_same(rte *x, rte *y)
 {
   /* rte.flags / rte.pflags are not checked, as they are internal to rtable */
@@ -1560,9 +1490,10 @@ rte_update_unlock(void)
 void
 rte_update2(struct channel *c, const net_addr *n, rte *new, struct rte_src *src)
 {
-  // struct proto *p = c->proto;
+  struct proto *p = c->proto;
   struct proto_stats *stats = &c->stats;
   const struct filter *filter = c->in_filter;
+  struct mpls_fec *fec = NULL;
   net *nn;
 
   ASSERT(c->channel_state == CS_UP);
@@ -1611,6 +1542,17 @@ rte_update2(struct channel *c, const net_addr *n, rte *new, struct rte_src *src)
 	    new->flags |= REF_FILTERED;
 	  }
 	}
+
+      if (p->mpls_map)
+        {
+	  if (mpls_handle_rte(p->mpls_map, n, new, rte_update_pool, &fec) < 0)
+	    {
+	      rte_trace_in(D_FILTERS, c, new, "invalid");
+	      stats->imp_updates_invalid++;
+	      goto drop;
+	    }
+	}
+
       if (!rta_is_cached(new->attrs)) /* Need to copy attributes */
 	new->attrs = rta_lookup(new->attrs);
       new->flags |= REF_COW;
@@ -1635,6 +1577,9 @@ rte_update2(struct channel *c, const net_addr *n, rte *new, struct rte_src *src)
   /* And recalculate the best route */
   rte_recalculate(c, nn, new, src);
 
+  if (p->mpls_map)
+    mpls_handle_rte_cleanup(p->mpls_map, &fec);
+
   rte_update_unlock();
   return;
 
@@ -1733,6 +1678,9 @@ rt_examine(rtable *t, net_addr *a, struct channel *c, const struct filter *filte
 void
 rt_refresh_begin(rtable *t, struct channel *c)
 {
+  if (c->debug & D_EVENTS)
+    log(L_TRACE "%s.%s: Route refresh begin", c->proto->name, c->name);
+
   FIB_WALK(&t->fib, net, n)
     {
       rte *e;
@@ -1754,6 +1702,9 @@ rt_refresh_begin(rtable *t, struct channel *c)
 void
 rt_refresh_end(rtable *t, struct channel *c)
 {
+  if (c->debug & D_EVENTS)
+    log(L_TRACE "%s.%s: Route refresh end", c->proto->name, c->name);
+
   int prune = 0;
 
   FIB_WALK(&t->fib, net, n)
@@ -1800,13 +1751,13 @@ rt_modify_stale(rtable *t, struct channel *c)
  * This functions dumps contents of a &rte to debug output.
  */
 void
-rte_dump(rte *e)
+rte_dump(struct dump_request *dreq, rte *e)
 {
   net *n = e->net;
-  debug("%-1N ", n->n.addr);
-  debug("PF=%02x ", e->pflags);
-  rta_dump(e->attrs);
-  debug("\n");
+  RDUMP("%-1N ", n->n.addr);
+  RDUMP("PF=%02x ", e->pflags);
+  rta_dump(dreq, e->attrs);
+  RDUMP("\n");
 }
 
 /**
@@ -1816,9 +1767,9 @@ rte_dump(rte *e)
  * This function dumps contents of a given routing table to debug output.
  */
 void
-rt_dump(rtable *t)
+rt_dump(struct dump_request *dreq, rtable *t)
 {
-  debug("Dump of routing table <%s>\n", t->name);
+  RDUMP("Dump of routing table <%s>\n", t->name);
 #ifdef DEBUGGING
   fib_check(&t->fib);
 #endif
@@ -1826,10 +1777,10 @@ rt_dump(rtable *t)
     {
       rte *e;
       for(e=n->routes; e; e=e->next)
-	rte_dump(e);
+	rte_dump(dreq, e);
     }
   FIB_WALK_END;
-  debug("\n");
+  RDUMP("\n");
 }
 
 /**
@@ -1838,13 +1789,13 @@ rt_dump(rtable *t)
  * This function dumps contents of all routing tables to debug output.
  */
 void
-rt_dump_all(void)
+rt_dump_all(struct dump_request *dreq)
 {
   rtable *t;
   node *n;
 
   WALK_LIST2(t, n, routing_tables, n)
-    rt_dump(t);
+    rt_dump(dreq, t);
 }
 
 static inline void
@@ -2106,10 +2057,10 @@ rt_free(resource *_r)
 }
 
 static void
-rt_res_dump(resource *_r)
+rt_res_dump(struct dump_request *dreq, resource *_r)
 {
   rtable *r = (rtable *) _r;
-  debug("name \"%s\", addr_type=%s, rt_count=%u, use_count=%d\n",
+  RDUMP("name \"%s\", addr_type=%s, rt_count=%u, use_count=%d\n",
       r->name, net_label[r->addr_type], r->rt_count, r->use_count);
 }
 
@@ -2133,6 +2084,7 @@ rt_setup(pool *pp, struct rtable_config *cf)
   t->name = cf->name;
   t->config = cf;
   t->addr_type = cf->addr_type;
+  t->debug = cf->debug;
 
   fib_init(&t->fib, p, t->addr_type, sizeof(net), OFFSETOF(net, n), 0, NULL);
 
@@ -2148,11 +2100,11 @@ rt_setup(pool *pp, struct rtable_config *cf)
   init_list(&t->flowspec_links);
   init_list(&t->subscribers);
 
+  hmap_init(&t->id_map, p, 1024);
+  hmap_set(&t->id_map, 0);
+
   if (!(t->internal = cf->internal))
   {
-    hmap_init(&t->id_map, p, 1024);
-    hmap_set(&t->id_map, 0);
-
     t->rt_event = ev_new_init(p, rt_event, t);
     t->prune_timer = tm_new_init(p, rt_prune_timer, t, 0, 0);
     t->last_rt_change = t->gc_time = current_time();
@@ -2409,8 +2361,8 @@ rt_preconfig(struct config *c)
 {
   init_list(&c->tables);
 
-  rt_new_table(cf_get_symbol("master4"), NET_IP4);
-  rt_new_table(cf_get_symbol("master6"), NET_IP6);
+  rt_new_table(cf_get_symbol(c, "master4"), NET_IP4);
+  rt_new_table(cf_get_symbol(c, "master6"), NET_IP6);
 }
 
 void
@@ -2838,13 +2790,14 @@ rt_new_table(struct symbol *s, uint addr_type)
 
   struct rtable_config *c = cfg_allocz(sizeof(struct rtable_config));
 
-  cf_define_symbol(s, SYM_TABLE, table, c);
+  cf_define_symbol(new_config, s, SYM_TABLE, table, c);
   c->name = s->name;
   c->addr_type = addr_type;
   c->gc_threshold = 1000;
   c->gc_period = (uint) -1;	/* set in rt_postconfig() */
   c->min_settle_time = 1 S;
   c->max_settle_time = 20 S;
+  c->debug = new_config->table_default_debug;
 
   add_tail(&new_config->tables, &c->n);
 
@@ -2902,6 +2855,7 @@ rt_reconfigure(rtable *tab, struct rtable_config *new, struct rtable_config *old
   new->table = tab;
   tab->name = new->name;
   tab->config = new;
+  tab->debug = new->debug;
 
   return 1;
 }
@@ -3096,9 +3050,6 @@ rte_update_in(struct channel *c, const net_addr *n, rte *new, struct rte_src *sr
 	{
 	  old->flags &= ~(REF_STALE | REF_DISCARD | REF_MODIFY);
 
-	  if (c->proto->rte_update_in_notify)
-	    c->proto->rte_update_in_notify(c, n, old, src);
-
 	  return 1;
 	}
 
@@ -3111,28 +3062,15 @@ rte_update_in(struct channel *c, const net_addr *n, rte *new, struct rte_src *sr
 
       /* Remove the old rte */
       *pos = old->next;
-      rte_free_quick(old);
       tab->rt_count--;
-
       break;
     }
 
-  if (!new)
-  {
-    if (!old)
-      goto drop_withdraw;
-
-    if (!net->routes)
-      fib_delete(&tab->fib, net);
-
-    if (c->proto->rte_update_in_notify)
-      c->proto->rte_update_in_notify(c, n, NULL, src);
-
-    return 1;
-  }
+  if (!old && !new)
+    goto drop_withdraw;
 
   struct channel_limit *l = &c->rx_limit;
-  if (l->action && !old)
+  if (l->action && !old && new)
   {
     if (tab->rt_count >= l->limit)
       channel_notify_limit(c, l, PLD_RX, tab->rt_count);
@@ -3147,18 +3085,39 @@ rte_update_in(struct channel *c, const net_addr *n, rte *new, struct rte_src *sr
     }
   }
 
-  /* Insert the new rte */
-  rte *e = rte_do_cow(new);
-  e->flags |= REF_COW;
-  e->net = net;
-  e->sender = c;
-  e->lastmod = current_time();
-  e->next = *pos;
-  *pos = e;
-  tab->rt_count++;
+  if (new)
+  {
+    /* Insert the new rte */
+    rte *e = rte_do_cow(new);
+    e->flags |= REF_COW;
+    e->net = net;
+    e->sender = c;
+    e->lastmod = current_time();
+    e->next = *pos;
+    *pos = new = e;
+    tab->rt_count++;
 
-  if (c->proto->rte_update_in_notify)
-    c->proto->rte_update_in_notify(c, n, e, src);
+    if (!old)
+    {
+      new->id = hmap_first_zero(&tab->id_map);
+      hmap_set(&tab->id_map, new->id);
+    }
+    else
+      new->id = old->id;
+  }
+
+  rte_announce(tab, RA_ANY, net, new, old, NULL, NULL);
+
+  if (old)
+  {
+    if (!new)
+      hmap_clear(&tab->id_map, old->id);
+
+    rte_free_quick(old);
+  }
+
+  if (!net->routes)
+    fib_delete(&tab->fib, net);
 
   return 1;
 
@@ -3643,6 +3602,7 @@ rt_get_hostentry(rtable *tab, ip_addr a, ip_addr ll, rtable *dep)
       return he;
 
   he = hc_new_hostentry(hc, tab->rp, a, link, dep, k);
+  he->owner = tab;
   rt_update_hostentry(tab, he);
   return he;
 }
